"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Object subclass:#Collection
	instanceVariableNames:''
	classVariableNames:'InvalidKeySignal EmptyCollectionSignal ValueNotFoundSignal
		NotEnoughElementsSignal'
	poolDictionaries:''
	category:'Collections-Abstract'
!

!Collection class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

documentation
"
    Abstract superclass for all collections.
    This abstract class provides functionality common to all collections,
    without knowing how the concrete class implements things. Thus, all
    methods found here depend on some basic mechanisms to be defined in the
    concrete class. 
    These basic methods are usually defined as #subclassResponsibility here.
    Some methods are also redefined for better performance.
"
! !

!Collection class methodsFor:'initialization'!

initialize
    "setup the signal"

    InvalidKeySignal isNil ifTrue:[
	InvalidKeySignal := ErrorSignal newSignalMayProceed:true.
	InvalidKeySignal nameClass:self message:#invalidKeySignal.
	InvalidKeySignal notifierString:'invalid key:'.

	ValueNotFoundSignal := ErrorSignal newSignalMayProceed:true.
	ValueNotFoundSignal nameClass:self message:#valueNotFoundSignal.
	ValueNotFoundSignal notifierString:'value not found:'.

	NotEnoughElementsSignal := ErrorSignal newSignalMayProceed:true.
	NotEnoughElementsSignal nameClass:self message:#notEnoughElementsSignal.
	NotEnoughElementsSignal notifierString:'not enough elements in collection'.

	EmptyCollectionSignal := NotEnoughElementsSignal newSignalMayProceed:true.
	EmptyCollectionSignal nameClass:self message:#emptyCollectionSignal.
	EmptyCollectionSignal notifierString:'operation not allowed for empty collections'.
    ]
! !

!Collection class methodsFor:'instance creation'!

new:size withAll:element
    "return a new COllection of size, where all elements are
     initialized to element"

    |newCollection|

    newCollection := self new:size.
    size timesRepeat:[newCollection add:element]
!

with:anObject
    "return a new Collection with one element:anObject"

    |newCollection|

    newCollection := self new.
    newCollection add:anObject.
    ^ newCollection
!

with:firstObject with:secondObject
    "return a new Collection with two elements:firstObject and secondObject"

    |newCollection|

    newCollection := self new.
    newCollection add:firstObject; add:secondObject.
    ^ newCollection
!

with:firstObject with:secondObject with:thirdObject
    "return a new Collection with three elements"

    |newCollection|

    newCollection := self new.
    newCollection add:firstObject; add:secondObject; add:thirdObject.
    ^ newCollection
!

with:firstObject with:secondObject with:thirdObject with:fourthObject
    "return a new Collection with four elements"

    |newCollection|

    newCollection := self new.
    newCollection add:firstObject; add:secondObject; add:thirdObject; add:fourthObject.
    ^ newCollection
!

with:one with:two with:three with:four with:five
    "return a new Collection with five elements"

    |newCollection|

    newCollection := self new.
    newCollection add:one; add:two; add:three; add:four; add:five.
    ^ newCollection
!

withAll:aCollection
    "return a new Collection with all elements taken from the argument,
     aCollection"

    |newCollection|

    newCollection := self new.
    newCollection addAll:aCollection.
    ^newCollection
! !

!Collection class methodsFor:'Signal constants'!

emptyCollectionSignal
    "return the signal used to report non-allowed operation on empty collections"

    ^ EmptyCollectionSignal
!

invalidKeySignal
    "return the signal used to report bad key usage"

    ^ InvalidKeySignal
!

notEnoughElementsSignal
    "return the signal used to report attempts for an operation, for which
     there are not enough elements in the collection"

    ^ NotEnoughElementsSignal
!

valueNotFoundSignal
    "return the signal used to report a nonexisting element."

    ^ ValueNotFoundSignal
! !

!Collection class methodsFor:'queries'!

growIsCheap
    "return true, if this collection can easily grow
     (i.e. without a need for become:).
     Returns true here; this method is redefined in fix-size
     collections"

    ^ true
! !

!Collection methodsFor:'accessing'!

anElement
    "return any element from the collection, or nil if there is none"

    self do: [:each | ^ each].
    ^ nil
!

first
    "return the first element of the collection.
     This should be redefined in subclasses."

    self do:[:e | ^ e].

    "error if collection is empty"
    ^ self emptyCollectionError
!

firstIfEmpty:exceptionValue
    "return the first element of the collection.
     If its empty, return the excveptionValue.
     (i.e. dont trigger an error as done in #first)"

    self isEmpty ifTrue:[^ exceptionValue value].
    ^ self first

    "Modified: 6.2.1996 / 15:27:17 / cg"
!

last
    "return the last element of the collection.
     This should be redefined in subclasses."

    |theLastOne any|

    any := false.
    self do:[:e | any := true. theLastOne := e].
    any ifTrue:[
	^ theLastOne
    ].

    "error if collection is empty"
    ^ self emptyCollectionError
!

lastIfEmpty:exceptionValue
    "return the last element of the collection.
     If its empty, return the exceptionValue.
     (i.e. dont trigger an error as done in #last)"

    self isEmpty ifTrue:[^ exceptionValue value].
    ^ self last

    "Modified: 6.2.1996 / 15:27:17 / cg"
    "Created: 6.2.1996 / 15:27:49 / cg"
!

upTo:anElement
    "return a new collection consisting of the receivers elements upTo
     (but excluding) anElement.
     If anElement is not in the receiver, the returned collection
     will consist of all elements of the receiver"

    |s|

    s := ReadStream on:self.
    ^ s upTo:anElement

    "
     'test.text' upTo:$.  
     #(1 2 3 4 5) upTo:4  
    "
! !

!Collection methodsFor:'adding & removing'!

add:anObject
    "add the argument, anObject to the receiver.
     If the receiver is ordered, the position of the new element is undefined
     (i.e. dont depend on where it will be put)."

    ^ self subclassResponsibility
!

addAll:aCollection
    "add all elements of the argument, aCollection to the receiver.
     Returns the argument, aCollection."

    aCollection do:[:element |
        self add:element
    ].
    ^ aCollection

    "
     #(1 2 3 4) copy addAll:#(5 6 7 8)
     #(1 2 3 4) asOrderedCollection addAll:#(5 6 7 8)
    "

    "Modified: 12.4.1996 / 13:29:20 / cg"
!

addAllFirst:aCollection
    "insert all elements of the argument, aCollection at the beginning
     of the receiver. Returns the argument, aCollection."

    aCollection reverseDo:[:element | 
        self addFirst:element 
    ].
    ^ aCollection

    "
     |c|
     c := #(4 3 2 1) asOrderedCollection.
     c addAllFirst:#(9 8 7 6 5).
     c   
    "

    "Modified: 12.4.1996 / 13:30:10 / cg"
!

addAllLast:aCollection
    "add all elements of the argument, aCollection to the receiver.
     Returns the argument, aCollection."

    aCollection do:[:element | 
        self addLast:element 
    ].
    ^ aCollection

    "
     |c|
     c := #(4 3 2 1) asOrderedCollection.
     c addAllLast:#(9 8 7 6 5)
    "

    "Modified: 12.4.1996 / 13:30:54 / cg"
!

addFirst:anObject
    "add the argument, anObject to the receiver.
     If the receiver is ordered, the new element will be added at the beginning."

    ^ self subclassResponsibility
!

addLast:anObject
    "add the argument, anObject to the receiver. 
     If the receiver is ordered, the new element will be added at the end.
     Return the argument, anObject."

    ^ self add:anObject

    "Modified: 12.4.1996 / 13:31:12 / cg"
!

remove:anObject
    "remove the (first occurrence of) argument, anObject from the receiver,
     and return it.
     If it was not in the collection, raise an error."

    ^ self remove:anObject ifAbsent:[self errorValueNotFound:anObject]

    "Modified: 12.4.1996 / 13:33:30 / cg"
!

remove:anObject ifAbsent:exceptionBlock
    "remove the (first occurrence of) argument, anObject from the receiver,
     and return it.
     If it was not in the collection, return the the value of the exceptionBlock"

    ^ self subclassResponsibility
!

removeAll
    "remove all elements from the receiver. Returns the receiver.
     This should be reimplemented in subclasses for better
     performance."

    [self notEmpty] whileTrue:[
        self removeFirst
    ].

    "Modified: 12.4.1996 / 13:34:39 / cg"
!

removeAll:aCollection
    "remove all elements of the argument, aCollection from the receiver.
     Return the argument, aCollection.

     Notice: for some collections (those not tuned for
             resizing themself) this may be very slow.
             If the number of removed elements is big compared to to
             the receivers size, it may be better to copy the
             ones which are not to be removed into a new collection."

    aCollection do:[:element | self remove:element].
    ^ aCollection

    "Modified: 12.4.1996 / 13:35:48 / cg"
!

removeFirst
    "remove the first element from the receiver.
     Return the removed element."

    ^ self subclassResponsibility
!

removeFirst:n
    "remove the first n elements from the receiver.
     Return an array filled with the removed elements.
     Notice: for some collections (those not tuned for
	     resizing themself) this may be very slow."

    |ret|

    self size < n ifTrue:[
	^ self notEnoughElementsError
    ].
    ret := Array new:n.
    1 to:n do:[:i |
	ret at:i put:(self removeFirst).
    ].
    ^ ret
!

removeLast
    "remove the last element from the receiver.
     Return the removed element."

    ^ self subclassResponsibility
!

removeLast:n
    "remove the last n elements from the receiver.
     Return an array filled with the removed elements.
     Notice: for some collections (those not tuned for
	     resizing themself) this may be very slow."

    |ret|

    self size < n ifTrue:[
	^ self notEnoughElementsError
    ].
    ret := Array new:n.
    n to:1 by:-1 do:[:i |
	ret at:i put:(self removeLast).
    ].
    ^ ret
! !

!Collection methodsFor:'converting'!

asArray
    "return a new Array with the collections elements"

    |anArray 
     index "{ Class: SmallInteger }" |

    anArray := Array new:(self size).
    index := 1.
    self do:[:each |
	anArray at:index put:each.
	index := index + 1
    ].
    ^ anArray
!

asBag
    "return a new Bag with the receiver collections elements"

    ^ self addAllTo:(Bag new)
!

asByteArray
    "return a new ByteArray with the collections elements
     (which must convert to integers in the range 0..255)."

    |aByteArray 
     index "{ Class: SmallInteger }" |

    aByteArray := ByteArray new:(self size).
    index := 1.
    self do:[:each |
	aByteArray at:index put:each asInteger.
	index := index + 1
    ].
    ^ aByteArray
!

asDoubleArray
    "return a new DoubleArray with the collections elements
     (which must convert to floats)."

    |aDoubleArray
     index "{ Class: SmallInteger }" |

    aDoubleArray := DoubleArray new:(self size).
    index := 1.
    self do:[:each |
	aDoubleArray at:index put:each asFloat.
	index := index + 1
    ].
    ^ aDoubleArray
!

asFloatArray
    "return a new FloatArray with the collections elements
     (which must convert to floats)."

    |aFloatArray
     index "{ Class: SmallInteger }" |

    aFloatArray := FloatArray new:(self size).
    index := 1.
    self do:[:each |
	aFloatArray at:index put:each asFloat.
	index := index + 1
    ].
    ^ aFloatArray
!

asIdentitySet
    "return a new IdentitySet with the receiver collections elements"

    ^ self addAllTo:(IdentitySet new:self size)
!

asOrderedCollection
    "return a new OrderedCollection with the receiver collections elements"

    ^ self addAllTo:(OrderedCollection new:self size)
!

asSet
    "return a new Set with the receiver collections elements"

    ^ self addAllTo:(Set new:self size)
!

asSortedCollection
    "return a new SortedCollection with the receiver collections elements"

    |aSortedCollection|

    aSortedCollection := SortedCollection new:self size.
    aSortedCollection addAll:self.
    ^ aSortedCollection
!

asSortedCollection:sortBlock
    "return a new SortedCollection with the receiver collections elements,
     using sortBlock for comparing"

    |aSortedCollection|

    aSortedCollection := SortedCollection sortBlock:sortBlock.
    aSortedCollection addAll:self.
    ^ aSortedCollection
!

asString
    "return a String with the collections elements 
     (which must convert to characters)"

    |aString 
     index "{ Class: SmallInteger }" |

    aString := String new:(self size).
    index := 1.
    self do:[:each |
	aString at:index put:each asCharacter.
	index := index + 1
    ].
    ^ aString
!

asStringCollection
    "return a new Text-object (collection of lines) with the elements printstings"

    ^ self addAllTo:(StringCollection new)
!

asText
    "return a Text-object (collection of lines) from myself.
     BIG warning: asText is totally misnamed here 
     - ST/X's asText has nothing to do with PP's asText.
     Therefore it will be removed/renamed soon. 
     Please use #asStringCollection."

    self obsoleteMethodWarning:'use #asStringCollection; #asText will change its meaning'.

    ^ self asStringCollection
!

readStream
    "return a stream for reading from the receiver"

    ^ ReadStream on:self

    "
     |s|

     s := 'hello world' readStream.
     s next:5.
     s next.
     (s next:5) inspect
    "
!

writeStream
    "return a stream for writing onto the receiver"

    ^ WriteStream on:self

    "
     |s|

     s := #() copy writeStream.
     s nextPut:1.
     s nextPut:2.
     s nextPut:3.
     s contents inspect
    "

    "
     |s|

     s := OrderedCollection new writeStream.
     s nextPut:1.
     s nextPut:2.
     s nextPut:3.
     s contents inspect
    "
! !

!Collection methodsFor:'copying'!

copyEmpty
    "return a copy of the receiver with no elements.
     This is used by copying and enumeration methods
     to get a new instance which is similar to the receiver."

    ^ self species new
!

copyEmpty:size
    "return a copy of the receiver with no elements, but space for
     size elements. This is used by copying and enumeration methods
     to get a new instance which is similar to the receiver.
     This method should be redefined in subclasses with instance
     variables, which should be put into the copy too.
     For example, SortedCollection has to copy its sortBlock into the
     new collection."

    ^ self species new:size
!

copyEmptyAndGrow:size
    "return a copy of the receiver with size nil elements.
     This is used by copying and enumeration methods
     to get a new instance which is similar to the receiver."

    ^ (self copyEmpty:size) grow:size
! !

!Collection methodsFor:'enumerating'!

addAllTo:aCollection
    "add all elements of the receiver, to aCollection.
     Return aCollection."

    self do:[:each | aCollection add:each].
    ^ aCollection
!

collect:aBlock
    "for each element in the receiver, evaluate the argument, aBlock
     and return a new collection with the results"

    |newCollection|

    newCollection := self species new.
    self do:[:element | newCollection add:(aBlock value:element)].
    ^ newCollection

    "
     #(1 2 3 4) collect:[:n | n * 2]  
    "
!

conform:aBlock
    "return true, if every element conforms to some condition.
     I.e. return false, if aBlock returns false for any element;
     true otherwise."

    self do:[:element | (aBlock value:element) ifFalse:[^ false]].
    ^ true
!

count:aBlock
    "count elements, for which aBlock returns true.
     Return the sum."

    ^ self inject:0 into:[:sumSoFar :element |
	    (aBlock value:element) ifTrue:[sumSoFar + 1]
				   ifFalse:[sumSoFar]
      ]

    "
     #(1 2 3 4 6 8 10) count:[:a | a even]     
     #(1 nil nil nil 2 3 nil 4 5) count:[:a | a isNil] 
    "
!

detect:aBlock
    "evaluate the argument, aBlock for each element in the receiver until
     the block returns true; in this case return the element which caused
     the true evaluation.
     If none of the evaluations returns true, report an error"

    ^ self detect:aBlock ifNone:[self errorNotFound]

    "
     #(1 2 3 4) detect:[:n | n odd]   
     #(2 4 6 8) detect:[:n | n odd]  
    "
!

detect:aBlock ifNone:exceptionBlock
    "evaluate the argument, aBlock for each element in the receiver until
     the block returns true; in this case return the element which caused
     the true evaluation.
     If none of the evaluations returns true, return the result of the
     evaluation of the exceptionBlock"

    self do:[:each | 
	(aBlock value:each) ifTrue:[^ each].
    ].
    ^ exceptionBlock value

    "
     #(1 2 3 4) detect:[:n | n odd] ifNone:['sorry']    
     #(2 4 6 8) detect:[:n | n odd] ifNone:['sorry']     
    "
!

detectLast:aBlock
    "evaluate the argument, aBlock for each element in the receiver until
     the block returns true; in this case return the element which caused
     the true evaluation. The elements are processed in reverse order.
     If none of the evaluations returns true, report an error"

    ^ self detectLast:aBlock ifNone:[self errorNotFound]

    "
     #(1 2 3 4) detectLast:[:n | n odd]   
     #(2 4 6 8) detectLast:[:n | n odd]  
    "
!

detectLast:aBlock ifNone:exceptionBlock
    "evaluate the argument, aBlock for each element in the receiver until
     the block returns true; in this case return the element which caused
     the true evaluation. The elements are processed in reverse order.
     If none of the evaluations returns true, return the result of the
     evaluation of the exceptionBlock"

    self reverseDo:[:each | 
	(aBlock value:each) ifTrue:[^ each].
    ].
    ^ exceptionBlock value

    "
     #(1 2 3 4) detectLast:[:n | n odd] ifNone:['sorry']    
     #(2 4 6 8) detectLast:[:n | n odd] ifNone:['sorry']     
    "
!

do:aBlock
    "evaluate the argument, aBlock for each element"

    ^ self subclassResponsibility
!

doWithExit:aBlock
    "evaluate the argument, aBlock for each element.
     Passes an additional exit object, which can be used to leave
     the loop early, by sending it a #value: message.
     Returns nil or the value passed to the exit>>value: message.

     Notice, that this is different to a return from the block, which
     returns from the enclosed method, NOT only from the block."

    ^ [:exit |
        self do:[:el |
            aBlock value:el value:exit
        ].
        nil
    ] valueWithExit.

    "Example:
     search for the first element which is >= 99; return it.
     remaining elements are not processed:

     #(1 2 3 4 999 5 6 7 8 9) 
        doWithExit:[:element :exit |
            Transcript showCr:element.
            element >= 99 ifTrue:[exit value:element]]

     #(1 2 3 4 5 6 7 8 9) 
        doWithExit:[:element :exit |
            Transcript showCr:element.
            element >= 99 ifTrue:[exit value:element]]
    "

    "to demonstrate the difference to returning from the block:
     this works:

     |el|

     el := #(1 2 3 4 999 5 6 7 8 9) 
        doWithExit:[:element :exit |
            element >= 99 ifTrue:[exit value:element]].
     Transcript showCr:el.


     this does NOT work as expected by a newComer ;-) (the showCr is not reached):

     |el|

     el := #(1 2 3 4 999 5 6 7 8 9) 
        do:[:element |
            element >= 99 ifTrue:[^ element]].
     Transcript showCr:el.
    "

    "Modified: 18.4.1996 / 14:16:59 / cg"
!

inject:thisValue into:binaryBlock
    |nextValue|

    nextValue := thisValue.
    self do: [:each | nextValue := binaryBlock value:nextValue value:each].
    ^ nextValue

    "
     #(1 2 3 4) inject:0 into:[:accu :element | accu + element]   
     (1 to:10) inject:0 into:[:accu :element | accu + element]     
    "
!

keysAndValuesDo:aTwoArgBlock
    "evaluate the argument, aBlock for every element in the collection,
     passing both index and element as arguments."

    ^ self errorNotKeyed
!

pairsDo:aBlock
    "evaluate the argument, aBlock for every element in the collection,
     which is supposed to consist of 2-element collections.
     The block is called with 2 arguments for each collection in the receiver."

    self do:[:aPair |
	aBlock value:(aPair at:1) value:(aPair at:2).
    ]
    "
     #( (1 one) (2 two) (3 three) (4 four) (5 five) (6 six)) 
     pairsDo:[:num :sym | Transcript show:num; show:' is: '; showCr:sym]


     #( (1 1)  (1 2)  (1 3)  (1 4)  (1 5)) 
     pairsDo:[:x :y | Transcript showCr:x@y]
    "
!

reject:aBlock
    "return a new collection with all elements from the receiver, for which
     the argument aBlock evaluates to false"

    ^ self select:[:element | (aBlock value:element) == false]

    "
     #(1 2 3 4) reject:[:e | e odd]   
     (1 to:10) reject:[:e | e even]     
    "
!

reverseDo:aBlock
    "evaluate the argument, aBlock for each element in reverse order."

    "it could be defined in terms of do: - but very inefficient.
     Better force programmer to define a better version ..."

    ^ self subclassResponsibility
!

select:aBlock
    "return a new collection with all elements from the receiver, for which
     the argument aBlock evaluates to true"

    |newCollection|

    newCollection := self species new.
    self do:[:each |
	(aBlock value:each) ifTrue:[newCollection add:each].
    ].
    ^ newCollection

    "
     #(1 2 3 4) select:[:e | e odd]   
     (1 to:10) select:[:e | e even]     
    "
!

select:aBlock ifNone:exceptionBlock
    "try a new collection with all elements from the receiver, for which
     the argument aBlock evaluates to true. If none of the elements passes
     the check of aBlock, return the result of evaluating exceptionBlock."

    |newCollection|

    newCollection := self select:aBlock.
    newCollection isEmpty ifTrue:[^ exceptionBlock value].
    ^ newCollection

    "
     #(1 2 3 4) select:[:e | e > 10] ifNone:['sorry']  
    "
!

triplesDo:aBlock
    "evaluate the argument, aBlock for every element in the collection,
     which is supposed to consist of 3-element collections.
     The block is called with 3 arguments for each collection in the receiver."

    self do:[:aTriple |
	aBlock value:(aTriple at:1) value:(aTriple at:2) value:(aTriple at:3).
    ]
    "
     #( (1 one eins) (2 two zwei) (3 three drei) (4 four vier) (5 five #'fnf') (6 six sechs)) 
     triplesDo:[:num :sym1 :sym2 | Transcript show:num; space; show:sym1; space; showCr:sym2]
    "
!

with:aCollection collect:aTwoArgBlock
    "evaluate the argument, aBlock for successive elements from
     each the receiver and the argument, aSequenceableCollection;
     The second argument, aBlock must be a two-argument block, which is
     evaluated for each element-pair.
     Collect the results and return a collection containing them.
     This method fails if neither the receiver nor aCollection is
     a sequenceable collection (i.e. implements numeric key access)."

    |index  "{ Class: SmallInteger }" 
     newCollection|

    newCollection := self species new.
    index := 1.
    aCollection isSequenceable ifFalse:[
	self isSequenceable ifFalse:[
	    ^ self error:'neither collection is sequenceable'.
	].
	aCollection do:[:element |
	    newCollection add:(aTwoArgBlock value:(self at:index) 
					    value:element).
	    index := index + 1
	]
    ] ifTrue:[
	self do:[:element |
	    newCollection add:(aTwoArgBlock value:element 
					    value:(aCollection at:index)).
	    index := index + 1
	]
    ].
    ^ newCollection

    "
     (1 to:3) with:#(one two three) collect:[:num :sym | (num->sym)]
     #(1 2 3) with:#(10 20 30) collect:[:x :y | (x@y)]
    "
!

with:aCollection do:aTwoArgBlock
    "evaluate the argument, aBlock for successive elements from
     each the receiver and the argument, aSequenceableCollection.
     The second argument, aBlock must be a two-argument block.
     This method fails if neither the receiver nor aCollection is
     a sequenceable collection (i.e. implements numeric key access)."

    |index  "{ Class: SmallInteger }"|

    index := 1.
    aCollection isSequenceable ifFalse:[
	self isSequenceable ifFalse:[
	    ^ self error:'neither collection is sequenceable'.
	].
	aCollection do:[:element |
	    aTwoArgBlock value:(self at:index) value:element.
	    index := index + 1
	]
    ] ifTrue:[
	self do:[:element |
	    aTwoArgBlock value:element value:(aCollection at:index).
	    index := index + 1
	]
    ]

    "
     (1 to:3) with:#(one two three) do:[:num :sym |
	Transcript showCr:(num->sym)
     ]
     (1 to:3) with:#(one two three) asSet do:[:num :sym |
	Transcript showCr:(num->sym)
     ].
     (1 to:3) asSet with:#(one two three) do:[:num :sym |
	Transcript showCr:(num->sym)
     ]
    "
! !

!Collection methodsFor:'error handling'!

emptyCheck 
    "check if the receiver is empty; report an error if so"

    self size == 0 ifTrue:[
	^ self emptyCollectionError.
    ].
!

emptyCollectionError
    "report an error that the operation is not allowed for
     empty collections"

    ^ EmptyCollectionSignal raise
!

errorInvalidKey:aKey
    "report an error that the given key was invalid"

    ^ InvalidKeySignal raiseRequestWith:aKey
!

errorNotKeyed
    "report an error that keyed access methods are not allowed"

    ^ self error:(self class name, 's do not respond to keyed accessing messages')
!

errorValueNotFound:anObject
    "report an error that an object was not found in the collection"

    ^ ValueNotFoundSignal raiseRequestWith:anObject
!

notEnoughElementsError
    "report an error that the operation is not allowed,  
     since not enough elements are in the collection"

    ^ NotEnoughElementsSignal raise
! !

!Collection methodsFor:'growing'!

grow
    "make the receiver larger"

    self grow:(self size + self growSize)
!

grow:howBig
    "change the receivers size"

    ^ self subclassResponsibility
!

growSize
    "return a suitable size increment for growing.
     The default returned here may be (and is) redefined in subclasses."

    ^ self size max:2
! !

!Collection methodsFor:'printing & storing'!

displayString
    "return a printed representation of the receiver for display in inspectors etc."

    ^ self printOrDisplayStringUsing:#displayString 
!

maxPrint
    "the print-limit; printOn: will try to not produce more output
     than the limit defined here."

    ^ 5000
!

printOn:aStream
    "append a user readable representation of the receiver to aStream.
     The text appended is not meant to be read back for reconstruction of
     the receiver. Also, this method limits the size of generated string.
    "

    |limit firstOne string|

    thisContext isRecursive ifTrue:[
	Transcript showCr:'Error: printOn: of self referencing collection.'.
	aStream nextPutAll:'#("recursive")'.
	^ self
    ].

    limit := self maxPrint.
    aStream nextPutAll:self class name.
    aStream nextPut:$(.
    firstOne := true.

    self do:[:element |
	firstOne ifFalse:[
	    aStream space
	] ifTrue:[
	    firstOne := false
	].
	(limit <= 0) ifTrue:[
	    aStream nextPutAll:'...etc...)'.
	    ^ self
	] ifFalse:[
"/ old code, worked only on positionable streams
"/
"/              pos1 := aStream position.
"/              element printOn:aStream.
"/              limit := limit - (aStream position - pos1)

	    string := element printString.
	    aStream nextPutAll:string.
	    limit := limit - string size.
	].
    ].
    aStream nextPut:$)

    "
     #(1 2 3 'hello' $a) printOn:Transcript
     (Array new:100000) printOn:Transcript
     (Array new:100000) printString size 
     (Dictionary new at:#hello put:'world'; yourself) printOn:Transcript
    "
    "
     |a| 
     a := Array new:3. 
     a at:2 put:a.
     a printOn:Transcript
    "
!

printOrDisplayStringUsing:aSelector
    "common code for printString and displayString; they only differ in
     the print-message sent to the elements"

    |thisString buffer count string noneYet total limit|

    thisContext isRecursive ifTrue:[
	Transcript showCr:'Error: print/storeString of self referencing collection.'.
	^ '#("recursive")'
    ].

    string := (self class name) , '('.
    noneYet := true.
    buffer := ''.
    count := 0.
    total := 0.
    limit := self maxPrint.

    self do: [:element |
	thisString := element perform:aSelector.
	noneYet ifTrue:[
	    noneYet := false.
	    buffer := buffer , thisString
	] ifFalse:[
	    buffer := buffer , (' ' , thisString)
	].
	count := count + 1.
	(count == 20) ifTrue:[
	    string := string , buffer.
	    buffer := ''.
	    count := 0
	].
	total := total + 1.
	(total > limit) ifTrue:[
	    string := string , buffer , '... )'.
	    ^ string
	]
    ].
    string := string , buffer , ')'.
    ^ string
!

storeOn:aStream
    "output a printed representation onto the argument, aStream.
     The text can be re-read to reconstruct (a copy of) the receiver.
     Recursive (i.e. cyclic) collections cannot be stored correctly
     (use storeBinaryOn: to handle those)."

    |isEmpty|

    thisContext isRecursive ifTrue:[
	Transcript showCr:'Error: storeOn: of self referencing collection.'.
	aStream nextPutAll:'#recursive'.
	^ self
    ].

    aStream nextPutAll:'('.
    aStream nextPutAll:(self class name).
    aStream nextPutAll:' new'.
    isEmpty := true.
    self do:[:element |
	aStream nextPutAll:' add:'.
	element storeOn:aStream.
	aStream nextPutAll:';'.
	isEmpty := false
    ].
    isEmpty ifFalse:[aStream nextPutAll:' yourself'].
    aStream nextPut:$)

    "
     OrderedCollection new storeOn:Transcript
     (1 to:10) storeOn:Transcript
     (Set new add:1; add:'hello'; yourself) storeOn:Transcript
    "
    "
     |s|

     s := Set new.
     s add:1; add:'hello'; add:s.
     s storeOn:Transcript
    "
! !

!Collection methodsFor:'queries'!

isCollection
    "return true, if the receiver is some kind of collection;
     true is returned here - the method is redefined from Object."

    ^ true
!

isSorted
    "return true, if the receiver is sorted.
     Collections which hold their elements in sorted order
     should return true. Some algorithms (quicksort) degenerate when 
     operating on sorted collections and can be avoided if this information
     is given. The default returned here (false) should not hurt.
     I.e. you should NEVER depend on that in your application."

    ^ false

    "Created: 13.4.1996 / 12:35:55 / cg"
!

longestCommonPrefix
    |try allMatching matchLen|

    "
     find the longest common prefix
    "
    matchLen := 0.
    self do:[:aName |
	aName size > matchLen ifTrue:[
	    matchLen := aName size.
	    try := aName
	]
    ].

    allMatching := true.

    [true] whileTrue:[
	allMatching := true.
	self do:[:aName |
	    (aName startsWith:try) ifFalse:[
		allMatching := false
	    ]
	].
	allMatching ifTrue:[
	    ^ try
	].
	matchLen := matchLen - 1.
	try := try copyTo:matchLen.
    ].
    ^ try

    "
     #('Array' 'ArrayedCollection' 'ArrayOfFoo') longestCommonPrefix 
     #('Arra' 'ArrayedCollection' 'ArrayOfFoo') longestCommonPrefix 
     #('Arra' 'b' 'c') longestCommonPrefix 
     #( (1 2 3 4) (1 2 3) (1 2 3 7) (1 2 3 9 10 11)) longestCommonPrefix
    "
! !

!Collection methodsFor:'testing'!

capacity
    "return the number of elements, that the receiver is
     prepared to take. For most collections, this is the actual
     size. However, some have more space preallocated to allow
     for faster adding of elements. 
     Not used by the system; added for ST-80 compatibility."

    ^ self size
!

includes:anElement
    "return true, if an object equal to the argument, anObject is in the list.
     This compares using #= (i.e. it does not look for the object itself,
     instead, one that compares equal).
     See #includesIdentical: when identity is asked for."

    self do:[:element |
	(anElement = element) ifTrue:[^ true].
    ].
    ^ false
!

includesAll:aCollection
    "return true, if the the receiver includes all elements of
     the argument, aCollection; false if any is missing.
     Notice: this method has O-square runtime behavior and may be
	     slow for big receivers/args. Think about using a Set,
	     or Dictionary."

    aCollection do:[:element |
	(self includes:element) ifFalse:[^ false].
    ].
    ^ true

    "
     #(1 2 3 4 5 6 7) includesAll:#(1 2 3)
     #('hello' 'there' 'world') includesAll:#('hello' 'world')
     #(1 2 3 4 5 6 7) includesAll:#(7 8 9)
    "
!

includesAny:aCollection
    "return true, if the the receiver includes any elements of
     the argument, aCollection; false if it includes none.
     Notice: this method has O-square runtime behavior and may be
	     slow for big receivers/args. Think about using a Set,
	     or Dictionary. Speedup is possible, by arrangy highly
	     probable elements towards the beginning of aCollection,
	     to avoid useless searches."

    aCollection do:[:element |
	(self includes:element) ifTrue:[^ true].
    ].
    ^ false

    "
     #(1 2 3 4 5 6 7) includesAny:#(1 2 3)
     #('hello' 'there' 'world') includesAny:#('hello' 'world')
     #(1 2 3 4 5 6 7) includesAny:#(7 8 9)
     #(1 2 3 4 5 6 7) includesAny:#(8 9 10)
    "
!

includesIdentical:anElement
    "return true, if the argument, anObject is in the list.
     This compares using #== (i.e. object identity).
     See #includes: when equality is asked for."

    self do:[:element |
	(anElement == element) ifTrue:[^ true].
    ].
    ^ false
!

isEmpty
    "return true, if the receiver is empty"

    ^ self size == 0
!

max
    "return the maximum value in the receiver collection"

    ^ self inject:nil
	    into:[:maxSoFar :this | 
			(maxSoFar isNil 
			or:[this > maxSoFar]) ifTrue:[this]
			ifFalse:[maxSoFar]]

    "
     #(15 1 -9 10 5) max  
     (1 to:15) max  
    "
!

min
    "return the minimum value in the receiver collection"

    ^ self inject:nil
	    into:[:minSoFar :this | 
			(minSoFar isNil 
			or:[this < minSoFar]) ifTrue:[this]
			ifFalse:[minSoFar]]

    "
     #(15 1 -9 10 5) min 
     (1 to:15) min        
    "
!

notEmpty
    "return true, if the receiver is not empty"

    ^ self isEmpty not
!

occurrencesOf:anElement
    "return the number of occurrences of the argument, anElement in
     the receiver"

    |count "{ Class: SmallInteger }" |

    count := 0.
    self do:[:element |
	(anElement = element) ifTrue:[
	    count := count + 1
	].
    ].
    ^ count
!

size
    "return the number of elements in the receiver.
     This is usually redefined in subclasses for more performance."

    |count "{ Class: SmallInteger }" |

    count := 0.
    self do:[:element |
	count := count + 1
    ].
    ^ count
! !

!Collection class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libbasic/Attic/Coll.st,v 1.49 1996-04-18 12:18:53 cg Exp $'
! !
Collection initialize!
